#include "upsd.h"
#include "serial.h"
#include <stdlib.h>

UPS_DRIVER("apcsmart", apcsmart, "APC Smart UPS serial driver")

/* in principle, after entering smart mode (or even in dumb mode!)
 * and asking for status bits, it should be sufficient to just
 * read data from UPS and intepret special chars like !$%+?=#&
 * (UPS sends those by its own when something changes).  But we
 * want to ensure we're still able to communicate with the UPS...
 */

static int apcsmart_poll(struct ups *ups) {
  char b[21];
  int r, s;
  r = serial_cmd(ups, "Q", 1, b, sizeof(b) - 1);
  if (r < 0)
    return r;
  if ((b[0] < '0' || b[0] >= '9') &&
      (b[0] < 'a' || b[0] >= 'f') &&
      (b[0] < 'A' || b[0] >= 'F'))
    return UPSR_PROTO;
  s = strtol(b, 0, 16);
  r = 0;
  /* 0x01 Calibration
     0x02 SmartTrim
     0x04 SmartBoost
     0x08 On Line
     0x10 on battery
     0x20 overload
     0x40 low battery
     0x80 replace battery
   */
  if (!(s & 0x08) || (s & 0x10)) r |= UPSS_ONBATT; /*XXX?*/
  if (s & 0x20) r |= UPSS_OVERLOAD;
  if (s & 0x40) r |= UPSS_LOWBATT;
  if (s & 0x80) r |= UPSS_BATTFAIL;
  return r;
}

static int apcsmart_cmd(struct ups *ups, enum ups_cmd cmd) {
  switch(cmd) {
  case UPSC_RESTART:
  case UPSC_HALT:
    /*XXX Kill power: either restore or shut off */
    return serial_write(ups, "S@000", 5);
  /*XXX case UPSC_SELFTEST: /o implement selftest for APCSmart */
  }
  return -1;
}

static int apcsmart_open(struct ups *ups, char **argv) {
  char b[256];
  int r, ntry, status;
  if (serial_open(ups, B2400, 1) < 0 ||
      serial_setmctrl(ups, TIOCM_DTR, TIOCM_RTS) < 0)
    return UPSR_ERROR;
  /* sometimes, an UPS is in interactive menu mode
   * (entered when \r character is sent to UPS).
   * There are several menu items and several submenu levels
   * (one is able to view and change various values and settings).
   * In this mode, UPS ignores smart-mode commands, but reacts to
   * characters displayed in the menu, while echoing anything else
   * back.  The menu mode have a timeout (2.5 minutes on our
   * "Smart-UPS OL 2000 XL"), but it is too large for our current
   * needs.  In order to go to the upper menu level, and eventually
   * back to normal mode, one have to send escape (033, 0x27 char)
   * to the UPS.  Here, we're trying to work around this situation
   * by first trying to go smart mode (Y) and if response is something
   * unexpecting, sending \033, eating all data from UPS, and
   * re-trying smart mode again.
   */
  ntry = 5;	/* max 5 levels to try */
  for(;;) {
    r = serial_cmd(ups, "Y", 1, b, sizeof(b)-1);
    if (r < 0) return r;
    if (b[0] == 'S' && b[1] == 'M')
      break;				/* success, smart mode */
    if (!ntry--)
      return UPSR_PROTO;		/* too many tries? */
    r = serial_cmd(ups, "\e", 1, b, sizeof(b)-1);
    if (!r)
      ntry = 0;
    else if (r > 0)
      do r = serial_read(ups, b, sizeof(b)-1);
      while(r > 0);
    if (r < 0)
      return r;
  }

  status = apcsmart_poll(ups);
  if (status < 0)
    return status;

  /* get the model string */
  r = serial_cmd(ups, "\1", 1, b, sizeof(b)-1);
  if (r < 0) return r;
  if (r < 3 || b[--r] != '\n' || b[--r] != '\r') r = 0;
  else while(r && b[r-1] == ' ') --r;
  b[r] = '\0';
  if (r == 2 && b[0] == 'N' && b[1] == 'A') r = 0;
  if (r) ups->model = estrdup(b);
  ups->pollfn = apcsmart_poll;
  ups->cmdfn = apcsmart_cmd;
  return status;
}
