#include "upsd.h"
#include "serial.h"
#include <stdio.h>		/* for sprintf() */

UPS_DRIVER("fenton", fenton, "Fenton serial driver (line proto)")

/* line-oriented protocol, as found in eg Powercom AdvanceIntelligent series.
 * Send \r-terminated command, read \r-terminated response.
 * Serial line setup: 2400 BPS, no parity.
 *
 * Serial cable (9 pins female D-type connector):
 *  COMPUTER                      UPS
 * =======================================
 * RX   <----------------------  TX  (pin 9)
 * TX   ---------------------->  RX  (pin6)
 * GND  <----------------------  GND (pin 7)
 *
 * Protocol:
 *
 * Command: Q1\r - status inquiry.  UPS returns a line:
 *  (MMM.M NNN.N PPP.P QQQ RR.R S.SS TT.T hgfedcba\r
 * (note first character in a reply is "("), where:
 *  MMM.M - input voltage (decimal), volts
 *  NNN.N - input fault voltage (see below)
 *  PPP.P - output voltage (decimal), volts
 *  QQQ   - load percentage (percent of maximum current)
 *  RR.R  - input frequency, HZ
 *  SS.S  - (or S.S) battery voltage (deimal), volts.
 *    For charge percentage, see rating info - F\r command below.
 *  TT.T  - temperature, C
 *  hgfedcba - bits (ascii '0' or '1' each):
 *   a - beeper on/off
 *   b - shutdown active
 *   c - test in progress
 *   d - UPS type in progress (?)
 *   e - UPS failed
 *   f - bypass/boost or buck active
 *   g - battery low
 *   h - utility fail.
 * Example:
 *  (208.4 140.0 208.4 034 59.9 2.05 35.0 00110000\r
 *   ivolt fvolt ovolt l%  freq batv temp hgfedcba (for reference)
 * means: input voltage is 208.4V, input fault voltage 140.0V,
 * output voltage 208.4V, current load 34%, input frequency is 59.9HZ,
 * battery voltage 2.05V, temperature 35.0C;
 * UPS failed (e=1), bypass/boost or buck (f=1).
 *
 * Command T\r : test for 10sec.
 *
 * Command TL\r : test until battery low.
 *
 * Command T<n>\r : test for n minutes, n is a number ranging from 01 to 99.
 * eg T03\r - test for 3 minutes.
 *
 * Command Q\r : toggle beeper.
 *
 * Command: S<n>\r : shutdown in n minutes, n is a number .1,.2,...,01,..,10,
 * eg S.5\r - shutdown after 30 sec (0.5 min).
 * According to Powercom docs, UPS will restore power after 10 sec when AC
 * returns.  But what's the difference with the next command when?
 *
 * Command S<n>R<m>\r : shutdown and restore.
 * Shutdown after n minutes (.1,.2,...,01,02,...,10), and return power after
 * m minutes (0001..9999), but only if AC is present (UPS will wait for AC
 * before powering on).
 *
 * Command C\r : cancel shutdown command (S<n>\r and S<n>R<m>\r).
 * If UPS is in shutdown waiting mode, the shutdown will be cancelled.
 * If UPS is in restore waitnig mode, it will be turned on immediately.
 *
 * Command: CT\r : cancel test in progress.
 *
 * Command: I\r : return UPS information.  UPS replies with
 *  #company_name UPS_model Version\r
 * (the "#" character is the part of reply).
 *  company_name is 15-character field, space-filled
 *  UPS_model is 10-character field, space-filled
 *  Version is 10-character filled, space-filled
 *
 * Command: F\r : request rating info.  UPS replies with
 *  #MMM.M QQQ SS.SS RR.R\r
 * ("#" is a part of reply), where:
 *  MMM.M - rating voltage
 *  QQQ   - rating current
 *  SS.SS (or SSS.S) - battery voltage
 *  RR.R  - frequency.
 * It looks like those are "nominal" values, so, e.g to calculate
 * battery charge level, one have to get current battery voltage from
 * Q1\r command and calculate a percentage from this SS.SS value,
 *  current_batt_voltage * 100 / nominal_batt_voltage
 * Note that QQQ in Q1\r command (load level) is in percents, not absolute.
 *
 * Command D\r : unknown.  Documentation says:
 *  Command Description
 *  D       Status inquiry *Disable
 *  Q1      Status inquiry
 *
 * Any unrecognized command will be echoed back.
 * unsupported values in responses are filled with "@" character.
 */

#define isdigit(x) ((x) >= '0' && (x) <= '9')
#define tonumber(x) ((x) - '0')

static char *getnum(char **sp, int exp) {
  char *s = *sp, *ss = s;
  if (!isdigit(*s)) return NULL;
  do ++s;
  while(isdigit(*s) || *s == '.');
  if (*s != exp) return NULL;
  *s++ = '\0';
  *sp = s;
  return ss;
}

static int num2d100(char *s) {
  int n = tonumber(*s++);
  while(isdigit(*s))
    n = n * 10 + tonumber(*s++);
  n *= 100;
  if (*s++ == '.' && isdigit(*s)) {
    n += tonumber(*s++) * 10;
    if (isdigit(*s))
      n += tonumber(*s);
  }
  return n;
}

#define rating_bvoltage  ds0

static int fenton_poll(struct ups *ups) {
  int r, i;
  char b[65];
  char *p, *v;
  static const unsigned char bitflags[8] = {
    UPSS_ONBATT, UPSS_LOWBATT, 0, UPSS_FAILURE, 0, UPSS_SELFTEST, 0, 0
  };
  r = serial_cmd(ups, "Q1\r", 3, b, sizeof(b) - 1);
  if (r < 0)
    return r;
  if (b[0] != '(' || b[r-1] != '\r') return UPSR_PROTO;
  p = b + 1;
#define gv(name) if (!(v = getnum(&p, ' '))) return UPSR_PROTO
#ifdef UPS_VARS
#define sv(name) gv(name); if (flags & UPSF_EXTINF) ups_setvar(st, name, v)
#else
#define sv(name) gv(name)
#endif
  sv("input.voltage");
  gv("fault.voltage");
  sv("output.voltage");
  sv("loadpct");
  sv("frequency");
  gv("batt.voltage");
#ifdef UPS_VARS
  if ((flags & UPSF_EXTINF) && ups->rating_bvoltage) {
    r = num2d100(v) * 100 / ups->rating_bvoltage;
    sprintf(b, "%d", r > 100 ? 100 : r);
    ups_setvar(st, "batt.chargepct", b);
  }
#endif
  sv("temperature");
#undef sv
#undef gv
  /* status "bits" */
  r = 0;
  for(i = 0; i < 8; ++i)
    if (p[i] == '1') r |= bitflags[i];
    else if (p[i] != '0') return UPSR_PROTO;
  if (p[8] != '\r') return UPSR_PROTO;
  return r;
}

static int fenton_cmd(struct ups *ups, enum ups_cmd cmd) {
  /*XXX according to other implementations,
   * this UPS also requires two different commands for shutdown,
   * depending on whenever it's online or on-battery.
   * I don't know whenever it's true -- don't have that UPS anymore...
   */
  char b[20];
  int n;
  switch(cmd) {
  case UPSC_RESTART:
  case UPSC_HALT:
    b[0] = 'S';
    if (UPS_DEFAULT_SHUTDOWN_DELAY < 60) {
      b[1] = '.'; b[2] = '0' + (UPS_DEFAULT_SHUTDOWN_DELAY * 10 / 60);
    }
    else if (UPS_DEFAULT_SHUTDOWN_DELAY < 600) {
      b[1] = '0'; b[2] = '0' + UPS_DEFAULT_SHUTDOWN_DELAY / 60;
    }
    else { b[1] = '1'; b[2] = '0'; }
    if (cmd == UPSC_HALT) n = 3;
    else { b[3] = b[4] = b[5] = '0'; b[6] = '1'; n = 7; }
    b[n++] = '\r';
    n = serial_write(ups, b, n);
    if (n <= 0)
      return n;
    serial_read(ups, b, sizeof(b));
    return 1;
  case UPSC_SELFTEST:
    return serial_write(ups, "T\r", 2) <= 0 ? -1 : 1;
  }
  return -1;
}

static int fenton_open(struct ups *ups, char **argv) {
  char b[65];
#if 0
  char *p;
#endif
  int r;

  if (serial_open(ups, B2400, 1) < 0 ||
      serial_setmctrl(ups, TIOCM_DTR, TIOCM_RTS) < 0)
    return UPSR_ERROR;

  /* reset state */
  if (serial_cmd(ups, "\r", 1, b, sizeof(b) - 1) < 0)
    return UPSR_ERROR;

#if 0
  /* read rating values */
  r = serial_cmd(ups, "F\r", 2, b, sizeof(b) - 1);
  if (r < 0)
    return r;
  p = b;
  if (*p++ != '#') return 0;
  if (!getnum(&p, ' ')) return 0;
  if (!getnum(&p, ' ')) return 0;
  if (!(p = getnum(&p, '\r'))) return 0;
  ups->rating_bvoltage = num2d100(p);	/* 1/100 V */
#endif

  r = fenton_poll(ups);
  if (r < 0)
    return r;

  /* get the model name etc */
  r = serial_cmd(ups, "I\r", 2, b, sizeof(b) - 1);
  if (r < 0)
    return r;
  else if (r) {
    /* #company_name UPS_model Version
     *  company_name is 15-character field, space-filled
     *  UPS_model is 10-character field, space-filled
     *  Version is 10-character filled, space-filled
     */
    if (b[0] != '#') return 0;
    do --r; while(b[r-1] == ' '); /* trim / *\r$/ */
    b[r] = '\0';
    if (r > 1)
      ups->model = estrdup(b+1);
  }

  ups->pollfn = fenton_poll;
  ups->cmdfn = fenton_cmd;
  return r;
}
