#include "upsd.h"
#include "serial.h"
#include <string.h>
#include <stdlib.h>
#include <math.h> /* for sqrt */

static const char pcm_help[] =
"Supported models:\n"
" Powercom King series (some models are determined automatically)\n"
" Powercom BlackKnight series (determined automatically)\n"
" Powercom Imperial series (determined automatically)\n"
"If the driver can't determine model and voltage, only basic status\n"
"will be monitored - no voltages, frequencies and load/battery levels;\n"
"shutdown command may also not work properly.\n"
"Options:\n"
" device=[serial:|usb:]/dev/port\n"
"  communication device to use.  Optional prefix (default is serial)\n"
"  gives type of communication device (USB is not supported yet)\n"
" model=kinVA|bntVA|imp|impVA|imd|imdVA|gen|tst[123]\n"
"  kin = King[PRO] models, bnt = BlackKnight, imp/imd = Imperial,\n"
"  gen to force generic (without additional status) UPS model\n"
"  tst1|2|3 to probe different voltage/load algo\n"
"  VA = wattage, such as kin425 or bnt801\n"
" voltage=100|110|115|120|200|220|230|240\n"
" shutdown=1|sec or shutdown=2|minsec\n"
"  type of shutdown sequence - one byte (sec) or two bytes (min:sec)\n"
;

UPS_DRIVER("pcm", pcm, "Powercom serial driver")

/* Powercom UPS.  There are at least 3 different protocols,
 * 2 of which are similar to each other (send one-byte command,
 * wait for binary reply), and 3rd is completely different,
 * line-oriented.
 *
 * First (binary) protocol is used in e.g. KingPRO and BlackKnightPRO
 * series.  The difference between the two variations is mostly the
 * amount of data returned by the status command - 11 bytes returned
 * by (some old) KingPRO and 16 bytes by (never) Kings, BlackKnight
 * and Imperial.  This module decodes the flags (ON_BATTERY, OVERLOADED
 * etc), which are the same in both cases, and optionally tries to
 * decode the rest (input/output voltage/frequency, load and battery level).
 *
 * Problem with decoding the levels is that (a) it all depends on the power
 * level (wattage) of a given UPS (which is determined as a table lookup)
 * and voltage, (b) there's little information about how all the values are
 * calculated, and (c) the info which is available contradicts with itself,
 * eg, a sample program from powercom uses completely different formulas and
 * coefficients than in their "specs" docs.  It's more: the available "specs"
 * contains some amount of errors and typos.
 *
 * You can experiment with different decodings available by giving various
 * model=XXX and voltage=YYY parameters for the driver.  Specifying
 * model=gen forces the driver to stop trying to interpret all the levels,
 * only basic status bits are handled (which are always handled).
 *
 * Only Imperial series are nice in this respect, giving ready-to-use values
 * without a need for decoding.
 *
 * Another problem is that different models expect different argument for
 * shutdown (killpower) command: for some, the argument is a single byte
 * indicating number of secounds before turning off, and for some, it
 * is 2 bytes, first being number of minutes and second - number of secounds.
 * Obviously, the two are incompatible, while the command used is the same.
 * The driver assumes that UPSes which return 11-byte (short) status expects
 * one-byte shutdown argument (sec), and those which return 16-byte status
 * expects 2-byte argument (min:sec).  You can change this using driver
 * parameter shutdown={1|2}.
 *
 * And second variant, line-oriented protocol, usually used for more
 * powerful (and expensive) models, isn't supported by this module.
 * It is implemented in, e.g. tupsman (google), and in drv_fenton.c
 * module in this upsd release.
 *
 */

/* Powercom binary protocol.  Send one-byte command (with possible
 * arguments, depending on the command), read optional binary reply.
 * Serial line setup: 1200 BPS, 8 data bits.
 *
 * Command 0x01 (1): request status info.
 * UPS replies with either 11 or 16 bytes of data (depending on model),
 * including status bits and various levels.  Note I wasn't able to figure
 * out which units are used for that levels (voltage, load level, frequency
 * etc), and those units are different for different models.
 * The data returned is as follows:
 *  byte 0 UPS load level
 *  byte 1 Battery charge level
 *  byte 2 Input voltage
 *  byte 3 Output voltage
 *  byte 4 Input frequency
 *  byte 5 product family line
 *    seen 0 (12 bytes status), 0x10(BNT?), 0x11, 0x80
 *  byte 6 Output frequency
 *  byte 7 UPS echo schedule count time high minute (0 - 255)
 *  byte 8 UPS echo schedule count time low minute (0 - 255)
 *  byte 9 statusA bits
 *   bit0 0x01 line fail (1 = inv, 0 = line)
 *   bit1 0x02 low battery (1 = battery low, 0 = normal)
 *   bit2 0x04 X
 *   bit3 0x08 AVR (1 = AVR, 0 = no AVR)
 *   bit4 0x10 AVR mode (1 = boost, 0 = buck)
 *   bit5 0x20 load status (1 = overload, 0 = normal)
 *   bit6 0x40 X
 *     From OptiUPS docs: 110/220V SYSTEM (1 = 110V, 0 = 220V)
 *   bit7 0x80 SD (shutdown delay?) mode
 *  byte 10 statusB bits
 *   bit0 0x01 UPS fault (1 = FAULT)
 *   bit1 0x02 battery status (1 = bad, 0 = normal)
 *   bit2 0x04 test mode (1 = test, 0, normal)
 *   bit3 0x08 buzzer silence (1 = actice)
 *   bit4 0x10 pre-SD count mode (1 = active)
 *   bit5 0x20 schedule count mode (1 = active)
 *   bit6 0x40 disable no load shutdown (1 = active)
 *   bit7 0x80 X
 *  byte 11 UPS model
 *  byte 12 UPS model sub-number (wattage, VA) and voltage
 *    low nibble (&0x0f) - voltage as index
 *    high nibble (&0xf0,>>4) - model number (power?) as index
 *  byte 13 is pre-SD count time min (0 - 255)
 *  byte 14 is pre-SD count time sec (0 - 59)
 *    time left before shutdown after 188
 *  byte 15 is X
 *    BNT uses this byte to calculate output voltage
 *
 * Command 0x03 (3): UPS self test
 *
 * Command 0x04 (4): Wakeup time setup, 2-byte argument:
 *  byte 1 is high min (0 - 255)
 *  byte 2 is low min (0 - 255)
 *   1 high min = 256 low min
 *   ex: 4 0 60 Set UPS schedule wakeup time 1 hr
 *
 * Command 0x05 (5): buzzer silence
 *
 * Command 0x0b (11): UPS wakeup
 *  (doesn't seem to work)
 *
 * Command 0x47 (71, 'G'): Enable "NO LOAD" shutdown
 * Command 0x67 (103, 'g'): Disable "NO LOAD" shutdown
 *   (see status byte 10, bit 6)
 *   Doesn't seem to do anything (the bit changes however)
 *
 * Command 0xb9 (185): Enable AC normal restart
 * Command 0xba (186): Disable AC normal restart
 *
 * Command 0xbc (188): shutdown (SD) after given delay,
 * possible restoring power after AC returns (depending
 * on command 0xb9 or 0xba commands).  Most problematic
 * command, because different UPS models expects different
 * incompatible arguments.  For example, some KingPRO models
 * expect one-byte argument (number of secounds before the
 * poweroff), while e.g. BlackKnight series expects 2-byte
 * argument, first byte is number of minutes and 2nd byte
 * is number of secounds before shutdown.  So, variant 1:
 *  1 byte argument, shutdown delay in secounds (0-255)
 * or variant 2:
 *  2 byte argument, shutdown delay:
 *   1st byte minutes (0-255)
 *   2nd byte secounds (0-59)
 *
 * Ex. 185 188 1 0 - shutdown after 1 minute, enabling
 * normal AC restart.
 *
 * If Byte5 == 0xff, it means "Imperial".  So far for IMP Byte11==0x4b(KIN).
 * The same protocol is used by OptiUPS, and here, Byte5=0xff,Byte11=0x4b
 * means it's "ES2 Series UPS", with the following model series (byte12 high
 * nibble):
 *   0x2, 0x3 = ES2-550C
 *   0x4 = ES2-800C (12V)
 *   0x5 = ES2-800C (24V)
 *   0x6, 0x7 = ES2-1000C
 *   0x8 = ES2-1500C
 *   0xA = ES2-2000C
 * Note that for OptiUPS there's an alternative "text" (line-based, ascii)
 * protocol available, -- only when byte5==0xff.  See
 * http://www.networkupstools.org/protocols/optiups.html for details.
 * This protocol doesn't seem to work with Powercom Imperial series UPSes.
 */

/* voltage tables.
 * We add 1 to the voltage index value (low nibble of byte12) read from UPS
 * to support 0 as "unknown".
 * There are 11 voltage levels, invalid ones (12..15) are mapped to 0.
 * Voltages between 120 and 200 are also invalid, but we keep them since
 * many powercom docs mentions them anyway.
 */
#define Nvoltages 12
static const unsigned char voltage[Nvoltages] =
 /*0   1   2   3   4 5 6 7   8   9  10  11 */
 { 0,100,110,115,120,0,0,0,200,220,230,240};
enum voltageI {
 V100=1, V110=2, V115=3, V120=4,
 V200=8, V220=9, V230=10, V240=11
};

/* models, index is high nibble of byte 12 */

enum KINmodelVAI {
  KIN0=0,KIN425,KIN500,KIN525,KIN625,KIN800,KIN1000,KIN1200,
  KIN1500,KIN1600,KIN2000,KIN2200,KIN2500,KIN3000,KIN5000};
static const unsigned short
 KINmodelVA[16] =
  {    0,   425,   500,   525,   625,   800,   1000,   1200,
     1500,   1600,   2000,   2200,   2500,   3000,    5000};

typedef unsigned char PCMloadD_t[2][Nvoltages];

static const PCMloadD_t
 /* load level divisor for loadpct_110_D().
  * Index by online/onbattery and voltage */
 /* 0 100 110 115 120   0   0   0 200 220 230 240 */
 KINloadD425 = {
   {0, 99, 88, 84, 80, 84, 84, 84, 86, 86, 81, 76},
   {0, 60, 54, 51, 48, 51, 51, 51, 53, 53, 50, 48}},
 KINloadD525 = {
   {0,127,113,106,100,106,106,106,109,109,103, 97},
   {0, 81, 72, 67, 62, 67, 67, 67, 65, 65, 62, 59}},
 KINloadD625 = {
   {0,131,115,107,103,107,107,107,110,110,105, 99},
   {0, 79, 70, 67, 64, 67, 67, 67, 65, 65, 61, 58}},
 KINloadD2Kp = {
   {0, 94, 94, 94, 94, 94, 94, 94,120,120,115,110},
   {0, 84, 77, 74, 70, 74, 74, 74, 77, 77, 74, 70}};
 /* 0 100 110 115 120   0   0   0 200 220 230 240 */

static const float
 KINr1X2v[] = {1,1,1,1,1.89,1.89,1.89,0.127,0.127,1.89,1.89,1.89,0.256},
 KINr1Y2v[] = {1.73,1.74,1.74,1.77,0.9,0.9,0.9,13.204,13.204,0.88,0.88,0.88,6.645},
 KINr1Z2v[] = {1.15,0.9,0.9,0.75,1.1,1.1,1.1,0.8,0.8,0.86,0.86,0.86,0.7};

enum BNTmodelVAI {
  BNT0=0,BNT400,BNT500,BNT600,BNT800,BNT801,BNT1000,BNT1200,BNT1500,BNT2000};
static const unsigned short
 BNTmodelVA[16] =
  {    0,   400,   500,   600,   800,   801,   1000,   1200,   1500,   2000};

static const PCMloadD_t BNTloadD[9] = {
 /* index by model-1 (excluding KIN0).
  * Note I don't have data for low-voltage UPSes, only 220,230 and 240V
  */
 /* 0 100 110 115 120   0   0   0 200 220 230 240 */
 {/*BNT400*/
   {0,  0,  0,  0,  0,  0,  0,  0,  0, 88, 83, 78}, /*400 801*/
   {0,  0,  0,  0,  0,  0,  0,  0,  0, 54, 52, 49}} /*400*/
,{/*BNT500*/
   {0,  0,  0,  0,  0,  0,  0,  0,  0,108,103, 98}, /*500 800 1000 1500 2000*/
   {0,  0,  0,  0,  0,  0,  0,  0,  0, 66, 64, 61}} /*500 800*/
,{/*BNT600*/
   {0,  0,  0,  0,  0,  0,  0,  0,  0,128,123,118}, /*600 1200*/
   {0,  0,  0,  0,  0,  0,  0,  0,  0, 86, 84, 81}} /*600*/
,{/*BNT800*/
   {0,  0,  0,  0,  0,  0,  0,  0,  0,108,103, 98}, /*500 800 1000 1500 2000*/
   {0,  0,  0,  0,  0,  0,  0,  0,  0, 66, 64, 61}} /*500 800*/
,{/*BNT801*/
   {0,  0,  0,  0,  0,  0,  0,  0,  0, 88, 83, 78}, /*400 801*/
   {0,  0,  0,  0,  0,  0,  0,  0,  0, 44, 42, 40}} /*801*/
,{/*BNT1000*/
   {0,  0,  0,  0,  0,  0,  0,  0,  0,108,103, 98}, /*500 800 1000 1500 2000*/
   {0,  0,  0,  0,  0,  0,  0,  0,  0, 56, 54, 52}} /*1000 1500 2000*/
,{/*BNT1200*/
   {0,  0,  0,  0,  0,  0,  0,  0,  0,128,123,118}, /*600 1200*/
   {0,  0,  0,  0,  0,  0,  0,  0,  0, 76, 74, 72}} /*1200*/
,{/*BNT1500*/
   {0,  0,  0,  0,  0,  0,  0,  0,  0,108,103, 98}, /*500 800 1000 1500 2000*/
   {0,  0,  0,  0,  0,  0,  0,  0,  0, 56, 54, 52}} /*1000 1500 2000*/
,{/*BNT2000*/
   {0,  0,  0,  0,  0,  0,  0,  0,  0,108,103, 98}, /*500 800 1000 1500 2000*/
   {0,  0,  0,  0,  0,  0,  0,  0,  0, 56, 54, 52}} /*1000 1500 2000*/
 /* 0 100 110 115 120   0   0   0 200 220 230 240 */
};


enum IMPmodelVAI                     /* unconfirmed 1025+ */
 {IMP0=0,IMP425,IMP525,IMP625,IMP825,IMP1025,IMP1200,IMP1500,IMP2000};
static const unsigned short IMPmodelVA[16] =
 {     0,   425,   525,   625,   825,   1025,   1200,   1500,   2000};

static const unsigned short TSTmodelVA[16] =
 { 0, 1, 2, 3 };

enum family {
  FAM_GEN, FAM_KIN, FAM_BNT, FAM_IMP, FAM_TST
};

static const char *PCMfam[] = {
  NULL, "King", "BlackKnight", "Imperial", "TEST"
};

static const unsigned short *PCMmodelVA[] = {
  NULL, KINmodelVA, BNTmodelVA, IMPmodelVA, TSTmodelVA
};

struct ups_driver_data {
  enum family fam;
  int ss;	/* size of status reply: 11 or 16 */
  unsigned ssb;	/* shutdown sequence bytes: 2=min:sec, 1=sec */
  unsigned VAi; /* VA index */
  unsigned Vi;  /* voltage index */
};

static int
battpct_0_29_100(unsigned bv, unsigned b0, unsigned b29, unsigned b100) {
  return
    bv <= b0   ? 0 :
    bv <= b29  ? (bv - b0) * 30000 / (b29 - b0) :
#if 1
/* don't cap to 100%, just to know actual value - which will tell us how
 * much real the data can be.. */
                 30000 + (bv - b29) * 70000 / (b100 - b29)
#else
    bv <= b100 ? 30000 + (bv - b29) * 70000 / (b100 - b29) :
                 100000
#endif
;
}

static int
loadpct_110_D(unsigned lv, unsigned D) {
  return D ? (int)(lv * 110000 / D) : -1;
}

static void freq4807(struct ups_cinfo *ci, const unsigned char *b) {
  #define _freq4807(v,x) if (x) v = 4807000 / x
  _freq4807(ci->ifreq, b[4]);
  _freq4807(ci->ofreq, b[6]);
  #undef _freq4807
}

static int
pcm_status(struct ups *ups, const unsigned char *b) {
  int r;
  const struct ups_driver_data *d = ups->data;
  struct ups_cinfo *ci = &ups->cinfo;
  int k;

//  if (b[7] != 0 || b[8] != 0)
//    return UPSR_PROTO;

  /* decode status bits */
  r = 0;
  if (b[9] & 0x01) r |= UPSS_ONBATT;
  if (b[9] & 0x02) r |= UPSS_LOWBATT;
  if (b[9] & 0x08) r |= b[9] & 0x10 ? UPSS_BOOST : UPSS_BUCK;
  if (b[9] & 0x20) r |= UPSS_OVERLOAD;
  if (b[10] & 0x01) r |= UPSS_FAILURE;
  if (b[10] & 0x02) r |= UPSS_BATTFAIL;
  if (b[10] & 0x04) r |= UPSS_SELFTEST;
  if (!b[3] && !b[6]) r |= UPSS_OFF;

  if (!(ups->flags & UPSF_CINFO))
    return r;

  /* decode other values (extended info) if/when known */

#define onbatt (r & UPSS_ONBATT)
#define avr (b[9] & 0x08)
#define avr_boost (b[9] & 0x10)
#define rawval(x) ((x) | 0x10000000u)

  switch(d->fam) {

  case FAM_GEN: {
    ci->loadpct = rawval(b[0]);
    ci->battpct = rawval(b[1]);
    ci->ivolt = rawval(b[2]);
    ci->ovolt = rawval(b[3]);
#if 1
    freq4807(ci,b);
#else
    ci->ifreq = rawval(b[4]);
    ci->ofreq = rawval(b[6]);
#endif
    break;
  }

  case FAM_KIN:
    freq4807(ci,b);

    switch((enum KINmodelVAI)d->VAi) {
    case KIN0: break;

    case KIN425:
    case KIN525:
    case KIN625:
      if (b[2])
        ci->ivolt = b[2] * 1790 + 3350; /*+*/
      if (b[3] && d->Vi) {
        if (!onbatt) {
          ci->ovolt =
            !avr ? b[3] * 1790 + 3350 : /*+*/
            avr_boost ? b[3] * 2070 :
                        b[3] * 1571;
        }
        else {
//const unsigned i = (b[12] / 16); // voltage from current data
const unsigned i = d->VAi-1;
const float x = KINr1X2v[i];
const float y = KINr1Y2v[i];
const float z = KINr1Z2v[i];
const unsigned t = b[6] / 2;
const float s = (b[1]*x - b[0]*z)*(b[1]*x - b[0]*z)*(t - b[3])/t;
printf("Vi=%d VAi=%d b[12]/16=%d\n", d->Vi, d->VAi, b[12]/16);
if (s > 0) ci->ovolt = sqrt(s) * y * 1000;
        }
      }
      ci->battpct = battpct_0_29_100(b[1] + b[0]/4, 169, 176, 204);
      { const PCMloadD_t *D;
        switch(d->VAi) {
        case KIN425: D = &KINloadD425; break;
        case KIN525: D = &KINloadD525; break;
        case KIN625: D = &KINloadD625; break;
        }
        ci->loadpct = loadpct_110_D(b[0], (*D)[onbatt?1:0][d->Vi]);
      }
      break;

    case KIN500: /*XXX in PCM specs KIN500 is in row of 425, 525, 525 above */

    case KIN800:
    case KIN1000:
    case KIN1200:
    case KIN1500:
    case KIN1600:
      if (b[2])
        ci->ivolt = b[2] * 1610;
      if (onbatt) {
        // ci->ovolt = ?
      }
      else
        ci->ovolt = b[3] * (
          !avr ? 1610 :
          avr_boost ? 1870 : 1370);

      k = (b[1] - 165) * 2600;
      if (onbatt)
        k += b[0] * 1000;
      else if (k > 7000)
        k -= 6000;
      ci->battpct = k;
 
      ci->loadpct = b[0] * 1130;
      break;

    case KIN2000:
    case KIN2200:
    case KIN2500:
    case KIN3000:
    case KIN5000:
      ci->ivolt = b[2] * 1625;
      ci->battpct = battpct_0_29_100(b[1] + b[0]/4 + b[0]/32, 175, 182, 209);
      ci->loadpct = loadpct_110_D(b[0], KINloadD2Kp[onbatt?1:0][d->Vi]);
      break;

    }
    break;

  case FAM_BNT:
    freq4807(ci,b);
    if (b[2])
      ci->ivolt = b[2] * 2200 - 24000; /* * 2.2 - 24 */
    if (!onbatt) {
      k = b[3] * 2200 - 24000; /* * 2.2 - 24 */
      if (k > 0)
        ci->ovolt =
         !avr      ? k :
         avr_boost ? k * 31 / 27 :
                     k * 27 / 31;
    } else { /* onbatt */
      if (b[6] /*ofreq*/) {
        const unsigned f = b[6] / 2;
        const float v = 1.965*b[15];
        const float t = v * v * (f - b[3]) / f;
        if (t > 0.0)
          ci->ovolt = sqrt(t) * 1000;
      }
    }
    ci->battpct = battpct_0_29_100(b[1] + b[0]/4, 157, 165, 193);
    if (d->Vi && d->VAi)
      ci->loadpct = loadpct_110_D(b[0], BNTloadD[d->VAi-1][onbatt?1:0][d->Vi]);
    break;

  case FAM_IMP:
    ci->ifreq = b[4] * 1000;
    ci->ofreq = b[6] * 1000;
    ci->ivolt = b[2]*2 * 1000;
    ci->ovolt = b[3]*2 * 1000;
    ci->battpct = b[1] * 1000;
    ci->loadpct = b[0] * 1000;
    break;

  case FAM_TST: /* test family */

    freq4807(ci, b);

    if (d->Vi) {
      if (b[2])
        ci->ivolt =
          b[2] * (d->VAi == 3 ? 1610 : 1900) / (d->Vi < V200 ? 2 : 1);
 
      if (!onbatt) {
        k = 10000 +
          !avr ? 0 :
          avr_boost ? (d->VAi == 3 ? -1430 : 1350) :
                      (d->VAi == 3 ?  1550 : 1418);
        k *= d->VAi == 3 ? 161 : 190;
        if (d->Vi < V200) k /= 2;
        ci->ovolt = b[3] * k / 100;
      }
      else {
        unsigned t = b[6] / 2;
        float k = (avr ? (d->VAi == 3 ? 0.11111 : 0.15) : 0.0) + 1;
        if (d->Vi >= V200)
          k *= d->VAi == 3 ? 0.9 : 1.03;
      }
    }
    /* load level */
    ci->loadpct = b[0] * (
      d->VAi == 1 ? 5000 :
      d->VAi == 2 ? 3200 :
      d->VAi == 3 ? (onbatt ? 1130 : 1660) :
      -1);

    /* batt level */
    k = (b[1] - 165) * (d->VAi == 1 ? 3100 : 2600);
    if (onbatt) k += b[0] * 1000;
    else if (k > 7000) k -= 6000;
    if (((r & UPSS_LOWBATT) && k > 29000) ||
        (!(r & UPSS_LOWBATT) && k < 29000))
      k = 29000;
    else if (k < 0) k = 0;
    else if (k > 100000) k = 100000;
    ci->battpct = k;
    break;

  }

#undef onbatt
#undef avr
#undef avr_boost

  return r;
}

/* send "status" command, read reply and decode status bits.
 * Return:
 *   <0 on error,
 *   0 if wrong/unrecognized data,
 *   or status bits if ok.
 * Note that this routine is used both by it's own (poll_fn),
 * and by powercomm_init() below.
 */

static int pcm_poll(struct ups *ups) {
  unsigned char b[20];
  int r;
  r = serial_cmd(ups, "\1", 1, b, sizeof(b));
  if (r < 0)	/* propagate the error condition */
    return r;
  if (r != ups->data->ss)
    return UPSR_PROTO;
  return pcm_status(ups, b);
}

static int pcm_cmd(struct ups *ups, enum ups_cmd cmd) {
  unsigned char b[5];
  int r;

  switch(cmd) {
  case UPSC_RESTART:
  case UPSC_HALT:
    r = 0;
    b[r++] = cmd == UPSC_RESTART ? 185 : 186;
    b[r++] = 188;
    switch(ups->data->ssb) {
    case 1: /* sec */
      b[r++] = UPS_DEFAULT_SHUTDOWN_DELAY > 255 ?
        255 : UPS_DEFAULT_SHUTDOWN_DELAY;
      break;
    case 2:
      b[r++] = UPS_DEFAULT_SHUTDOWN_DELAY / 60;
      b[r++] = UPS_DEFAULT_SHUTDOWN_DELAY % 60;
      break;
    default:
      fatal("shutdown sequence for this UPS is unknown");
    }
    if (serial_write(ups, b, r) < 0)
      return -1;
    //serial_read(ups, b, sizeof(b));	/* just for fun */
    return 1;

  case UPSC_SELFTEST:
    return serial_write(ups, "\003", 1) < 0 ? -1 : 1;

  case UPSC_WAKE:
    return serial_write(ups, "\013", 1) < 0 ? -1 : 1;
  }
  return -1;
}

static int s2bin(const char *s, unsigned char *b, unsigned bs) {
  unsigned char *be = b + bs;
  unsigned char *bp = b;
  unsigned n;
  while(*s && bp < be) {
    if (*s == '\\') {
      switch(*++s) {
      case '0': case '1': case '2': case '3':
      case '4': case '5': case '6': case '7':
        n = *s++ - '0';
        if (*s >= '0' && *s <= '7') n = n * 8 + *s++ - '0';
        if (*s >= '0' && *s <= '7') n = n * 8 + *s++ - '0';
        --s;
        break;
      case 't': n = '\t'; break;
      case 'n': n = '\n'; break;
      case 'r': n = '\r'; break;
      case 's': n = ' '; break;
      case 'v': n = '\v'; break;
      default: n = (unsigned char)*s;
      }
      *bp++ = n;
      ++s;
    }
    else
      *bp++ = *s++;
  }
  be = bp;
  printf("read %d:", be - b);
  for(bp = b; bp < be; ++bp)
    printf(" 0x%02x(%d)", *bp, *bp);
  putchar('\n');
  return be - b;
}

static int pcm_open(struct ups *ups, char **argv) {
  int r;
  unsigned char b[30];
  char name[60];
  int testdata = 0;
  int cmd = 0;

  /* model info: user specified and determined */
  int ufam = 0, uVi = 0, uVAi = 0, ussb = 0;
  int afam = 0, aVi = 0, aVAi = 0, assb = 0;

  struct ups_driver_data data;

  for(r = 0; argv[r]; ++r) {
    char *t = argv[r]; int x;
    if (strpfx(t, "model=")) {
      t += 6;
      if (strpfx(t, "kin")) ufam = FAM_KIN;
      else if (strpfx(t, "bnt")) ufam = FAM_BNT;
      else if (strpfx(t, "imp")) ufam = FAM_IMP;
      else if (strpfx(t, "imd")) ufam = FAM_IMP;
      else if (strpfx(t, "tst")) ufam = FAM_TST;
      else if (streq(t, "gen")) ufam = -1;
      else
        fatal("unknown powercom family `%s' specified for %s", t, ups->tag);
      t += 3;
      if (*t) {
        x = atoi(t);
        for(uVAi = 1; ; ++uVAi)
          if (uVAi >= 16)
            fatal("unknown wattage `%s' for %s", t, ups->tag);
          else if (PCMmodelVA[ufam][uVAi] == x && x)
            break;
      }
      else
        uVAi = 0;
    }
    else if (strpfx(t, "voltage=")) {
      t += 8;
      x = atoi(t);
      for(uVi = 1;; ++uVi)
        if (uVi >= Nvoltages)
          fatal("unknown voltage `%s' specified for %s", t, ups->tag);
        else if (voltage[uVi] == x && x)
          break;
    }
    else if (strpfx(t, "shutdown=")) {
      t += 9;
      if (streq2(t, "2", "minsec")) ussb = 2;
      else if (streq2(t, "1", "sec")) ussb = 1;
      else fatal("unknown shutdown argument `%s' for %s", t, ups->tag);
    }
    else if (strpfx(t, "testdata=")) {
      t += 9;
      testdata = s2bin(t, b, sizeof(b));
      cmd = 0;
    }
    else if (strpfx(t, "cmd=")) {
      t += 4;
      cmd = s2bin(t, b, sizeof(b));
      testdata = 0;
    }
    else
      fatal("unknown argument `%s' for %s", t, ups->tag);
  }

  if (strpfx(ups->device, "serial:")) ups->device += 7;
  else if (strpfx(ups->device, "com:")) ups->device += 4;
  else if (strpfx(ups->device, "usb:"))
   fatal("USB isn't supported yet by pcm driver");

  ups->flags |= UPSF_BINARY;

  if (!testdata) {
  if (serial_open(ups, B1200, 1) < 0)
    return UPSR_ERROR;

#if 0
  /* check status lines.  XXX which lines are up/down on ac/batt fail? */
  r = serial_getmctrl(ups);
  if (r < 0) return UPSR_ERROR;
  if (r) return UPSR_PROTO; /* some bits are set: probably not powercom... */
#endif

  /* according to NUT project, various models wants us to set RTS&DTR lines
   * differently.   Both RTS and DTR high seems to work.
   * Note that we use "raw" mode without any flow control.
   */
  if (serial_setmctrl(ups, TIOCM_RTS|TIOCM_DTR, 0) < 0)
    return UPSR_ERROR;
  if (cmd) {
    serial_cmd(ups, b, cmd, b, sizeof(b));
    return UPSR_ERROR;
  }
  r = serial_cmd(ups, "\1", 1, b, sizeof(b));
  if (r < 0) return r;
  }
  else r = testdata;

  memset(&data, 0, sizeof(data));

  data.ss = r;
  if (r == 11) {
    /* short status: can't determine wattage and voltage */
    dbg(1, "11-byte status: model=0x%02x", b[5]);
    assb = 1; /* assume one-byte shutdown sequence */
/* verified models:
 * King PRO 1200AP 220V RU b5=0x00 (bogus data for battpct)
 */
  }
  else if (r == 16) {
    /* full status, model, wattage and voltage are supported if known model */
    dbg(1, "16-byte status: b5=0x%02x%s b11=0x%02x%s b12=0x%02x",
        b[5], b[5] == 0xff ? "(imperial)" : "",
        b[11], b[11] == 0x4b ? "(king)" : b[11] == 0x4b ? "bnt" : "",
        b[12]);
    assb = 2; /* assume min:sec shutdown for all 16-byte status models */

    if (b[5] == 0xff) { afam = FAM_IMP;
    }
    else if (b[11] == 0x4b) { afam = FAM_KIN;
/* verified models:
 * 5=0x80, 11=0x4b, 12=0x38 = KIN525 220V RU
 */
    }
    else if (b[11] == 0x42) { afam = FAM_BNT;
/* verified models:
 * 5=0x10, 11=0x42, 12=0x28 = BlackKnight-500AP 220V RU
 */
    }

    if (afam) {
      r = b[12] >> 4;
      if (PCMmodelVA[afam][r])
        aVAi = r;
    }

    r = (b[12] & 0x0f) + 1;
    if (r < Nvoltages && voltage[r])
      aVi = r;
    else if (!uVi)
      warn("unable to determine voltage (raw=0x%x), specify voltage=NN",
           b[12] & 0x0f);

  }
  else {
    warn("PCM UPS should return either 11 or 16 bytes of status");
    return UPSR_PROTO;
  }

  //if (b[7] != 0 || b[8] != 0) return UPSR_PROTO;

  if (ufam < 0) /* special case for "generic" */
    ;
  else if (ufam) {
    data.fam = ufam;
    if (afam && ufam != afam)
      warn("actual UPS family (%s) differs from specified (%s)",
           PCMfam[afam], PCMfam[ufam]);
    data.VAi = uVAi;
    if (afam == ufam) {
      if (uVAi && aVAi && aVAi != uVAi)
        warn("actual UPS wattage (%dVA) differs from specified (%dVA)",
             PCMmodelVA[afam][aVAi], PCMmodelVA[ufam][uVAi]);
      if (!uVAi)
        data.VAi = aVAi;
    }
    /*XXX check bytes-to-read */
  }
  else if (!afam)
    warn("unable to determine UPS family (raw=0x%02x), specify model=XX", b[5]);
  else {
    data.fam = afam;
    data.VAi = aVAi;
  }

  if (!data.VAi && data.fam && data.fam != FAM_IMP)
    warn("unknown UPS wattage");

  if (uVi) {
    if (aVi && aVi != uVi)
      warn("actual UPS voltage (%dV) differs from specified (%dV)",
           voltage[aVi], voltage[uVi]);
    data.Vi = uVi;
  }
  else if (aVi)
    data.Vi = aVi;
  else if (data.fam && data.fam != FAM_IMP)
    warn("unable to determine UPS voltage, specify voltage=NN");

  data.ssb = ussb ? ussb : assb;

  if (data.fam) {
    char *s = name;
    s += sprintf(s, "%s", PCMfam[data.fam]);
    if (data.VAi)
      s += sprintf(s, "-%dAP", PCMmodelVA[data.fam][data.VAi]);
    if (data.Vi)
      s += sprintf(s, " %dV", voltage[data.Vi]);
  }
  else
    sprintf(name, "n=%d:f=0x%02x:m11=0x%02x:m12=0x%02x",
            r, b[5], r > 11 ? b[11] : 0, r > 12 ? b[12] : 0);

  if (data.fam == FAM_IMP || (data.fam && data.VAi && data.Vi))
    ups->flags |= UPSF_CINFO;

  r = 0;
  switch(data.fam) {
  case FAM_KIN:
    switch(data.VAi) {
    case KIN0:
      break; /* unknown */
    case KIN425:
    case KIN525:
    case KIN625:
      break; /* verified */
    case KIN500:
      r = 2; break; /* unsure */
    default:
      r = 1; /* need info */
    }
    break;
  case FAM_BNT:
    r = 1;
    break;
  case FAM_IMP:
    break;
  default:
    break;
  }
  if (r)
    warn("please confirm your UPS is handled OK - verify all the levels");

  ups->data = &data;
  r = pcm_status(ups, b);
  if (r < 0) return r;
  ups->pollfn = pcm_poll;
  ups->cmdfn = pcm_cmd;
  ups->model = estrdup(name);
  ups->data = emalloc(sizeof(data));
  memcpy(ups->data, &data, sizeof(data));
  return r;
}
