#include "upsd.h"
#include "serial.h"
#include <sys/ioctl.h>
#include <sys/file.h>
#include <limits.h>
#include <string.h>
#include <stdio.h>

static int
_eioctl(const struct ups *ups, int cmd, void *arg, const char *opname) {
  int r = ioctl(ups->fd, cmd, arg);
  if (r >= 0)
    return r;
  eerr("ups %s on %s failed", opname, ups->device);
  return UPSR_ERROR;
}
#define eioctl(ups, cmd, arg) _eioctl(ups, cmd, (void*)(arg), "ioctl(" #cmd ")")

#ifndef TIOCM_LE
#define TIOCM_LE	0x001
#endif
//#define TIOCM_DTR	0x002
//#define TIOCM_RTS	0x004
#ifndef TIOCM_ST
#define TIOCM_ST	0x008
#endif
#ifndef TIOCM_SR
#define TIOCM_SR	0x010
#endif
#define TIOCM_CTS	0x020
#ifndef TIOCM_CAR
#define TIOCM_CAR	0x040
#endif
#ifndef TIOCM_RNG
#warning TIOCM_RNG
#define TIOCM_RNG	0x080
#endif
#define TIOCM_DSR	0x100
#ifndef TIOCM_OUT1
#warning TIOCM_OUT1
#define TIOCM_OUT1	0x2000
#endif
#ifndef TIOCM_OUT2
#warning TIOCM_OUT2
#define TIOCM_OUT2	0x4000
#endif

void mctrl(const char *pfx, int f, const char *sfx) {
  fputs(pfx, stdout);
  if (!f)
    fputs(" (none)", stdout);
  else {
#define CB(B) \
	if (f & TIOCM_##B) fputs(" "#B, stdout), f &= ~TIOCM_##B
    CB(LE);
    CB(DTR);
    CB(RTS);
    CB(ST);
    CB(SR);
    CB(CTS);
    CB(CAR);
    CB(RNG);
    CB(DSR);
    CB(OUT1);
    CB(OUT2);
#undef CB
    if (f) printf(" 0x%x", f);
  }
  if (sfx) fputs(sfx, stdout);
}

/* performs modem control lines changes (TIOCM_RTS,TIOCM_DTR)
 */
int serial_setmctrl(struct ups *ups, int set, int clr) {
  if (debug) {
    mctrl("setting modem control lines:", set, 0);
    mctrl(", clearing:", clr, "\n");
  }
  if ((set && eioctl(ups, TIOCMBIS, &set) < 0) ||
      (clr && eioctl(ups, TIOCMBIC, &clr) < 0))
    return UPSR_ERROR;
  return 0;
}

/* obtains modem control lines status
 * (TIOCM_CD, TIOCM_CTS, TIOCM_DSR, TIOCM_RI)
 */
int serial_getmctrl(const struct ups *ups) {
  int r;
  if (eioctl(ups, TIOCMGET, &r) < 0)
    return UPSR_ERROR;
  if (debug)
    mctrl("modem control lines:", r, "\n");
  return r;
}

int serial_write(struct ups *ups, const void *wbuf, int wlen) {
  eioctl(ups, TCFLSH, TCIOFLUSH);
  return ups_write(ups, wbuf, wlen);
}

int serial_read(struct ups *ups, void *rbuf, int wlen) {
  return ups_read(ups, rbuf, wlen, 300);
}

int serial_cmd(struct ups *ups,
	       const void *wbuf, int wlen,
	       void *rbuf, int rlen) {
  wlen = serial_write(ups, wbuf, wlen);
  if (wlen < 0)
    return wlen;
  return serial_read(ups, rbuf, rlen);
}

/* opens the named device (serial port) and initializes the ups->tio
 * structure to be in "raw" mode
 */
int serial_open_tio(struct ups *ups, struct termio *tio) {
  ups->fd = open(ups->device, O_RDWR|O_NONBLOCK);
  if (ups->fd < 0) {
    ewarn("unable to open %s", ups->device);
    return UPSR_ERROR;
  }
  fcntl(ups->fd, F_SETFL, O_RDWR);
  if (flock(ups->fd, LOCK_EX|LOCK_NB) != 0) {
    ewarn("unable to lock %s", ups->device);
    close(ups->fd);
    ups->fd = -1;
    return UPSR_ERROR;
  }
  ups->openfn = 0;
  ups->pollfn = 0;
  ups->cmdfn = 0;
  if (eioctl(ups, TCFLSH, TCIOFLUSH) < 0 ||
      eioctl(ups, TCSETA, tio) < 0) {
    close(ups->fd);
    ups->fd = -1;
    return UPSR_ERROR;
  }
  return 0;
}

int serial_open(struct ups *ups, int speed, int raw) {
  struct termio tio;
  memset(&tio, 0, sizeof(tio));
  tio.c_cflag = speed | CS8 | CLOCAL | CREAD;
  tio.c_iflag = IGNPAR;
  tio.c_cc[VMIN] = MAX_INPUT;
  tio.c_cc[VTIME] = 1;
  return serial_open_tio(ups, &tio);
}
